@@ -1,5 +1,5 @@
 /*
-**  Copyright (c) 2012, The Trusted Domain Project.  All rights reserved.
+**  Copyright (c) 2012, 2013, The Trusted Domain Project.  All rights reserved.
 */
 
 #include "build-config.h"
@@ -98,7 +98,6 @@ struct dmarcf_header
 struct dmarcf_msgctx
 {
 	int			mctx_spfresult;
-	int			mctx_spfalign;
 	char *			mctx_jobid;
 	struct dmarcf_header *	mctx_hqhead;
 	struct dmarcf_header *	mctx_hqtail;
@@ -127,6 +126,7 @@ typedef struct dmarcf_connctx * DMARCF_CONNCTX;
 struct dmarcf_config
 {
 	_Bool			conf_afrf;
+	_Bool			conf_afrfnone;
 	_Bool			conf_rejectfail;
 	_Bool			conf_dolog;
 	_Bool			conf_enablecores;
@@ -412,6 +412,10 @@ dmarcf_parse_received_spf(char *str)
 				copying = FALSE;
 				parens++;
 			}
+ 			else if (isascii(*p) && isspace(*p))
+			{
+				copying = FALSE;
+			}
 			else if (r < end)
 			{
 				*r++ = *p;
@@ -481,6 +485,72 @@ dmarcf_addlist(const char *str, struct list **head)
 	}
 }
 
+/*
+**  DMARCF_LOAD_DNSDATA -- load fake DNS data into the library
+**
+**  Parameters:
+**  	path -- path to file to read
+**
+**  Return value:
+**  	FALSE iff load wasn't possible; caller should check errno
+*/
+
+_Bool
+dmarcf_load_dnsdata(char *path)
+{
+	_Bool gapfound;
+	const char *key;
+	const char *value;
+	char *p;
+	FILE *f;
+	char buf[BUFRSZ];
+
+	assert(path != NULL);
+
+	f = fopen(path, "r");
+	if (f == NULL)
+		return FALSE;
+
+	memset(buf, '\0', sizeof buf);
+
+	while (fgets(buf, sizeof buf - 1, f) != NULL)
+	{
+		key = NULL;
+		value = NULL;
+		gapfound = FALSE;
+
+		for (p = buf; *p != '\0'; p++)
+		{
+			if (*p == '\n' || *p == '#')
+			{
+				*p = '\0';
+				break;
+			}
+			else if (!isspace(*p))
+			{
+				if (key == NULL)
+					key = p;
+				else if (gapfound && value == NULL)
+					value = p;
+			}
+			else
+			{
+				if (!gapfound)
+				{
+					*p = '\0';
+					gapfound = TRUE;
+				}
+			}
+		}
+
+		opendmarc_dns_fake_record(key, value);
+	}
+
+	fclose(f);
+
+	return TRUE;
+}
+
 /*
 **  DMARCF_LOADLIST -- add a file's worth of entries to a singly-linked list
 **
@@ -1163,6 +1233,10 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		                  &conf->conf_afrf,
 		                  sizeof conf->conf_afrf);
 
+		(void) config_get(data, "ForensicReportsOnNone",
+		                  &conf->conf_afrfnone,
+		                  sizeof conf->conf_afrfnone);
+
 		(void) config_get(data, "RecordAllMessages",
 		                  &conf->conf_recordall,
 		                  sizeof conf->conf_recordall);
@@ -1195,6 +1269,19 @@ dmarcf_config_load(struct config *data, struct dmarcf_config *conf,
 		(void) config_get(data, "HistoryFile",
 		                  &conf->conf_historyfile,
 		                  sizeof conf->conf_historyfile);
+
+		str = NULL;
+		(void) config_get(data, "TestDNSData", &str, sizeof str);
+		if (str != NULL)
+		{
+			if (!dmarcf_load_dnsdata(str))
+			{
+				snprintf(err, errlen,
+				         "%s: can't load fake DNS data: %s",
+				         str, strerror(errno));
+				return -1;
+			}
+		}
 	}
 
 	if (conf->conf_authservid == NULL)
@@ -1549,7 +1636,7 @@ mlfi_connect(SMFICTX *ctx, char *host, _SOCK_ADDR *ip)
 	dmarcf_config_reload();
 
 	if (dmarcf_checkhost(host, ignore) ||
-	    dmarcf_checkip(ip, ignore))
+	    (ip != NULL && dmarcf_checkip(ip, ignore)))
 		return SMFIS_ACCEPT;
 
 	/* copy hostname and IP information to a connection context */
@@ -1690,6 +1777,7 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 	cc->cctx_msg = dfc;
 
 	dfc->mctx_jobid = JOBIDUNKNOWN;
+	dfc->mctx_spfresult = -1;
 
 	dfc->mctx_histbuf = dmarcf_dstring_new(BUFRSZ, 0);
 	if (dfc->mctx_histbuf == NULL)
@@ -1700,8 +1788,6 @@ mlfi_envfrom(SMFICTX *ctx, char **envfrom)
 		return SMFIS_TEMPFAIL;
 	}
 
-	dfc->mctx_spfalign = -1;
-
 	if (cc->cctx_dmarc != NULL)
 		(void) opendmarc_policy_connect_rset(cc->cctx_dmarc);
 
@@ -2019,6 +2105,7 @@ mlfi_eom(SMFICTX *ctx)
 				if (ar.ares_result[c].result_result != ARES_RESULT_PASS)
 					continue;
 
+				spfaddr = NULL;
 				spfmode = DMARC_POLICY_SPF_ORIGIN_HELO;
 
 				memset(addrbuf, '\0', sizeof addrbuf);
@@ -2051,6 +2138,18 @@ mlfi_eom(SMFICTX *ctx)
 					}
 				}
 
+				if (spfaddr == NULL)
+				{
+					if (conf->conf_dolog)
+					{
+						syslog(LOG_ERR,
+						       "%s: can't extract SPF address from Authentication-Results",
+						       dfc->mctx_jobid);
+					}
+
+					continue;
+				}
+
 				status = dmarcf_mail_parse(addrbuf, &user,
 				                           &domain);
 				if (status != 0 || domain == NULL ||
@@ -2145,10 +2244,25 @@ mlfi_eom(SMFICTX *ctx)
 		{
 			if (strcasecmp(hdr->hdr_name, RECEIVEDSPF) == 0)
 			{
-				dmarcf_dstring_printf(dfc->mctx_histbuf,
-				                      "spf %d\n",
-				                      dmarcf_parse_received_spf(hdr->hdr_value));
+				int spfres;
+				int spfmode;
 
+				if (dfc->mctx_fromdomain[0] == '\0')
+					spfmode = DMARC_POLICY_SPF_ORIGIN_HELO;
+				else
+					spfmode = DMARC_POLICY_SPF_ORIGIN_MAILFROM;
+
+				spfres = dmarcf_parse_received_spf(hdr->hdr_value);
+
+				dmarcf_dstring_printf(dfc->mctx_histbuf,
+				                      "spf %d\n", spfres);
+
+				/* use the MAIL FROM domain */
+				ostatus = opendmarc_policy_store_spf(cc->cctx_dmarc,
+				                                     dfc->mctx_envdomain,
+				                                     spfres,
+				                                     spfmode,
+				                                     NULL);
 				wspf = TRUE;
 			}
 		}
@@ -2257,7 +2371,8 @@ mlfi_eom(SMFICTX *ctx)
 
 	ruv = opendmarc_policy_fetch_ruf(cc->cctx_dmarc, NULL, 0, TRUE);
 	if ((policy == DMARC_POLICY_REJECT ||
-	     policy == DMARC_POLICY_QUARANTINE) &&
+	     policy == DMARC_POLICY_QUARANTINE ||
+	     (conf->conf_afrfnone && policy == DMARC_POLICY_NONE)) &&
 	    conf->conf_afrf && ruv != NULL)
 	{
 		_Bool first = TRUE;
@@ -2535,7 +2650,10 @@ mlfi_eom(SMFICTX *ctx)
 	}
 
 	if (conf->conf_dolog)
-		syslog(LOG_INFO, "%s: %s", dfc->mctx_jobid, aresult);
+	{
+		syslog(LOG_INFO, "%s: %s %s", dfc->mctx_jobid,
+		       dfc->mctx_fromdomain, aresult);
+	}
 
 	/* if the final action isn't TEMPFAIL or REJECT, add an A-R field */
 	if (ret != SMFIS_TEMPFAIL && ret != SMFIS_REJECT)
